package ldcontext

const (
	jsonPropMeta             = "_meta"
	jsonPropPrivate          = "privateAttributes"
	jsonPropRedacted         = "redactedAttributes"
	jsonPropOldUserCustom    = "custom"
	jsonPropOldUserPrivate   = "privateAttributeNames"
	jsonPropOldUserRedacted  = "privateAttrs"
	jsonPropOldUserSecondary = "secondary"
)
