// Package ldcontext defines the LaunchDarkly SDK model for contexts and context attributes.
//
// See [Context], [Builder], and [MultiBuilder].
//
// To learn more, read: https://docs.launchdarkly.com/home/contexts
package ldcontext
