// Package ldcommon is the base package for commonly used Go SDK types and functions.
//
// Currently there are no types in this package; they are all in subpackages such as ldvalue.
package ldcommon
