// Package ldlog contains a logging abstraction used by LaunchDarkly SDK components.
//
// These types and methods allow you to customize the logging behavior of the SDK, such as
// filtering log output by level or redirecting it to another destination.
//
// For a test implementation that captures output, see [github.com/launchdarkly/go-sdk-common/v3/ldlogtest].
package ldlog
