// Package sharedtest contains test helpers that may be used in unit tests and/or benchmarks for
// multiple packages in go-sdk-common. It should not import any go-sdk-common packages except ldvalue.
package sharedtest
